/**
 *  This file is part of DJFourierv4 developed by Blai Meléndez Catalán under
 	the supervision of prof. Horst Eidenberger (TU Wien).

    DJFourierv4 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DJFourierv4 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

 	<http://www.gnu.org/licenses/>. */

package com.example.djfourierv4;

public class FFT {

  int n, m;

  // Lookup tables. Only need to recompute when size of FFT changes.
  double[] cos;
  double[] sin;

  public FFT(int n) {
      this.n = n;
      this.m = (int) (Math.log(n) / Math.log(2));

      // Make sure n is a power of 2
      if (n != (1 << m))
          throw new RuntimeException("FFT length must be power of 2");

      // compute tables
      cos = new double[n / 2];
      sin = new double[n / 2];

      for (int i = 0; i < n / 2; i++) {
          cos[i] = Math.cos(-2 * Math.PI * i / n);
          sin[i] = Math.sin(-2 * Math.PI * i / n);
      }
  }

  public void fft(double[] x, double[] y) {
      int i, j, k, n1, n2, a;
      double c, s, t1, t2;

      // Bit-reverse
      j = 0;
      n2 = n / 2;
      for (i = 1; i < n - 1; i++) {
          n1 = n2;
          while (j >= n1) {
              j = j - n1;
              n1 = n1 / 2;
          }
          j = j + n1;

          if (i < j) {
              t1 = x[i];
              x[i] = x[j];
              x[j] = t1;
              t1 = y[i];
              y[i] = y[j];
              y[j] = t1;
          }
      }

      // FFT
      n1 = 0;
      n2 = 1;

      for (i = 0; i < m; i++) {
          n1 = n2;
          n2 = n2 + n2;
          a = 0;

          for (j = 0; j < n1; j++) {
              c = cos[a];
              s = sin[a];
              a += 1 << (m - i - 1);

              for (k = j; k < n; k = k + n2) {
                  t1 = c * x[k + n1] - s * y[k + n1];
                  t2 = s * x[k + n1] + c * y[k + n1];
                  x[k + n1] = x[k] - t1;
                  y[k + n1] = y[k] - t2;
                  x[k] = x[k] + t1;
                  y[k] = y[k] + t2;
              }
          }
      }	      
  }
  
  public void ifft(double[] x, double[] y, short[] xRes) {
      int i, j, k, n1, n2, a;
      double c, s, t1, t2;

      // Bit-reverse
      j = 0;
      n2 = n / 2;
      for (i = 1; i < n - 1; i++) {
          n1 = n2;
          while (j >= n1) {
              j = j - n1;
              n1 = n1 / 2;
          }
          j = j + n1;

          if (i < j) {
              t1 = x[i];
              x[i] = x[j];
              x[j] = t1;
              t1 = y[i];
              y[i] = y[j];
              y[j] = t1;
          }
      }

      // FFT
      n1 = 0;
      n2 = 1;

      for (i = 0; i < m; i++) {
          n1 = n2;
          n2 = n2 + n2;
          a = 0;

          for (j = 0; j < n1; j++) {
              c = cos[a];
              s = -sin[a];
              a += 1 << (m - i - 1);

              for (k = j; k < n; k = k + n2) {
                  t1 = c * x[k + n1] - s * y[k + n1];
                  t2 = s * x[k + n1] + c * y[k + n1];
                  x[k + n1] = x[k] - t1;
                  y[k + n1] = y[k] - t2;
                  x[k] = x[k] + t1;
                  y[k] = y[k] + t2;
              }
          }
      }
      
      //Amplitude control
      for (int d = 0; d < x.length; d++){
     	 
          x[d] = (x[d] /  n);
      }

	  double max = Tools.findMax(x)[0];
	  
	  if(max > 32768){
		  
		  x = Tools.normalizeShorts(x, max);
	  }
	  
	  for (int d = 0; d < x.length; d++){
	     	 
		  xRes[d] = (short) x[d];
      }
      
  }
  
  public void ifft(double[] x, double[] y, byte[] xRes) {
      int i, j, k, n1, n2, a;
      double c, s, t1, t2;

      // Bit-reverse
      j = 0;
      n2 = n / 2;
      for (i = 1; i < n - 1; i++) {
          n1 = n2;
          while (j >= n1) {
              j = j - n1;
              n1 = n1 / 2;
          }
          j = j + n1;

          if (i < j) {
              t1 = x[i];
              x[i] = x[j];
              x[j] = t1;
              t1 = y[i];
              y[i] = y[j];
              y[j] = t1;
          }
      }

      // FFT
      n1 = 0;
      n2 = 1;

      for (i = 0; i < m; i++) {
          n1 = n2;
          n2 = n2 + n2;
          a = 0;

          for (j = 0; j < n1; j++) {
              c = cos[a];
              s = -sin[a];
              a += 1 << (m - i - 1);

              for (k = j; k < n; k = k + n2) {
                  t1 = c * x[k + n1] - s * y[k + n1];
                  t2 = s * x[k + n1] + c * y[k + n1];
                  x[k + n1] = x[k] - t1;
                  y[k + n1] = y[k] - t2;
                  x[k] = x[k] + t1;
                  y[k] = y[k] + t2;
              }
          
          }
      }
      
      //Amplitude control
      for (int d = 0; d < x.length; d++){
     	 
          x[d] = (x[d] /  n);
      }

	  double max = Tools.findMax(x)[0];
	  
	  if(max > 128){
		  
		  x = Tools.normalizeBytes(x, max);
	  }
	  
	  for (int d = 0; d < x.length; d++){
	     	 
		  xRes[d] = (byte) x[d];
      }
  }
  
  /**
   * Calculates the FFT of two data pieces at a time */
  public void twoRealFFT(double dataReal1[], double dataImag1[], double dataReal2[], double dataImag2[]){
	  
	  double a, b, c, d;
	  
	  //Computation of the FFT.
	  fft(dataReal1, dataImag1);
	  
	  //Rearrangement of the values to separate the results of both FFTs.
	  dataReal2[0] = dataImag1[0];
	  dataImag1[0] = dataImag2[0] = 0;
	  
	  for(int j = 1; j <= n/2; j++){
		  
		  a = 0.5 * (dataReal1[j] + dataReal1[n - j]);
		  b = 0.5 * (dataReal1[j] - dataReal1[n - j]);
		  c = 0.5 * (dataImag1[j] + dataImag1[n - j]);
		  d = 0.5 * (dataImag1[j] - dataImag1[n - j]);
		  
		  dataReal1[j] = a;
		  dataImag1[j] = d;
		  dataReal1[n - j] = a;
		  dataImag1[n - j] = - d;
		  dataReal2[j] = c;
		  dataImag2[j] = - b;
		  dataReal2[n - j] = c;
		  dataImag2[n - j] = b;
	  }
  }
 }
